/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Matthew Wilson
 *    Dr. Katsuhiko Tsujino
 *    Eliot Setzer
 *    C. Greg Hagerty
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var iconManager = {};

function IconManager(annozilla0) {
    dump("IconManager is here\n");
    if (! IconManager._hasRequiredComponents()) {
        return;
    }

    this.annozilla = annozilla0;

    // whether we're inserting icons
    this.insertIcons = false;
    // whether we're inserting highlighting
    this.insertHighlight = false;

    // say that we are not observing annotations -- this must be done before loading prefs
    this.observing = false;

    // bundle
    this.bundle = srGetStrBundle("chrome://annozilla/locale/annozilla.properties");

    var that = this;

    // add a listener for when the window closes
    window.addEventListener("unload", function(e) { that.unload(); }, false);

    this.icons = {};

//  this.temporaryIcons = {};
}

// IconManager.prototype is a synonym for iconManager.
IconManager.prototype = iconManager;

//also used by ContextMenu.js
IconManager.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
IconManager.DC10_NS    = "http://purl.org/dc/elements/1.0/";
IconManager.TYPE_NS = "http://www.w3.org/2000/10/annotationType#";

/*******************
 * PUBLIC FUNCTIONS
 *******************/

IconManager.prototype.QueryInterface = function(iid) {  
  if (! (iid.equals(Components.interfaces.nsIAnnotationsObserver) ||
         iid.equals(Components.interfaces.nsISupports) ) ) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
  return this;
}

IconManager.prototype.unload = function() {
    if(this.observing){
        // remove observer
        this.annozilla.app.removeObserver(this);
        this.observing = false;
    }
}
  
// called whenever the insertIcons pref is set
IconManager.prototype.insertIconsPrefChanging = function(insert) {
    this.insertIcons = insert;
    this._insertPrefChange();
}

// called whenever the insertHighlight pref is set
IconManager.prototype.insertHighlightPrefChanging = function(insert) {
    this.insertHighlight = insert;
    this._insertPrefChange();
}

// called when either insertIcons or insertHighlight prefs are set 
IconManager.prototype._insertPrefChange = function() {
    var insert = this.insertIcons || this.insertHighlight;
    if (insert && (! this.observing) ) {
	// add our observer
        this.annozilla.app.addObserver(this);
	this.observing = true;
    }
    if ((!insert) && this.observing) {
	// remove our observer
	this.annozilla.app.removeObserver(this);
	this.observing = false;
    }
}  

/**
 * Called when a page begins loading in the 
 * browser if the application has 
 * autoLoad turned on.  
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url loading
 */
IconManager.prototype.autoLoadOnBeginPageLoad = function(uri, win) {}

/**
 * Called when a page successfully finishes loading.
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url finished loading
 * @param annotations nsISimpleEnumeration of annotations for url
 */
IconManager.prototype.autoLoadOnSuccessPageLoad = function(url, win) {
    //NOOP: wait for Annozilla to call us to load in order
}
/*
    // if this applies to another window, ignore
    if (window.content != win) {
	return;
    }

    if (this._isInsertIconsEnabled()) {

        //need to resolve annotations one at a time because each highlight 
        // changes DOM
        //resolving can take time it's better if we can do so asynchronously 
        this.annozilla.addAnnotations(this.annozilla.annotations);
    }
}
*/

/**
 * Called when a page that was loading is
 * unable to load (or is cancelled by the user).
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url aborted
 * @param message string error message
 */
IconManager.prototype.autoLoadOnFailPageLoad = function(url, win) {}


/**
 * Success callback function to return the results for an
 * automatic annotations load (autoLoad).
 * This function returns the annotations for a single server, designated by serverURI.  
 * It will be called once for each server that successfully responds.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI that
 *                    were on serverURI annotations server
 */
IconManager.prototype.autoLoadAnnotationsForServer = function(serverURI, win, annotatedURI, count, annotations) {
    //NOOP: wait for Annozilla to call us to load in order
}
/*
    // if this applies to another window, or
    // if there is a user load in progress, ignore these annotations
    if ((window.content != win) || this.annozilla.userLoadInProgress()) {
	return;
    }

    var state = this.annozilla.currentURIState();
    this._processAnnotationsWithAutoLoadInfo(state, annotations);
}
*/

/**
 * Callback function to indicate failure for an automatic
 * annotations load (autoLoad).
 * This function will be called once for each annotations server that
 * fails in the autoloading process. 
 *
 * @param serverURI string uri of the annotations server that failed
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */ 
IconManager.prototype.autoLoadFailureForServer = function(serverURI, win, annotatedURI, message) {}


/**
 * Success callback function to return the results for a loadAnnotations request
 *  - OR - an automatic autoLoad request.  This function returns the 
 * annotations for one server, designated by serverURI.  
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI that
 *                    were on serverURI annotations server
 */
IconManager.prototype.loadAnnotationsOnSuccess = function(serverURI, annotatedURI, annotationsLength, annotations) {

    window._content.document.loadBindingDocument("chrome://annozilla/content/bindings/annozillaBindings.xml");

    //NOOP: wait for Annozilla to call us to load in order
}
/*
    // if we're autoLoading, ignore these annotations
    if (! this.annozilla.userLoadInProgress()) {
        return;
    }
    
    // if we've moved on to a new page, ignore these annotations
    if (annotatedURI != this.annozilla.currentURI()) {
        return;
    }

    // This is now done in Annozilla.userLoadOnCommand.
//  this._removeAnnotations();  // Remove the existing annotation icons.

    // if autoLoad is enabled, we can benefit from its alerts about when the
    // page is fully loaded
    if (this.annozilla.isAutoLoadEnabled()) {
        this._processAnnotationsWithAutoLoadInfo(
            this.annozilla.currentURIState(),
            annotations
        );
    }
    // otherwise, we have to wing it. for now we'll assume that the page
    // is loaded, since there's no way that I know to check whether the DOM
    // is fully constructed
    else {
        var currentAnnotationsBatch = {};

        for (var i = 0; i < annotations.length; i++) {
            var current = annotations[i];

            // testing purposes
            if (! current.data) {
                this.annozilla._debugBadAnnotation(current);
                continue;
            }

            var range = this.annozilla.resolveAnnotation(current);

            // add to the current batch
            IconManager._addAnnotationToContainer(
                current,
                range,
                null,
                currentAnnotationsBatch
            );
        }

        this.annozilla.addAnnotations(currentAnnotationsBatch);
    }
}
*/

/**
 * Callback function to indicate failure for a loadAnnotations request
 *  - OR - an automatic autoLoad request.  This function indicates failure
 * for one particular annotations server, designated by serverURI.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */
IconManager.prototype.loadAnnotationsOnFailure = function(serverURI, annotatedURI, message) {}

/**
 * Called when an annotation post request succeeds.
 *
 * Only needs to do xpointer resolution if _isInsertIconsEnabled
 *   (Annozilla..postAnnotationOnSuccess performs xpointer resolution 
 *    independantly and redundantly in any case before adding to sidebar)
 *
 * @param newAnnotation nsIAnnotation the newly created annotation
 */
IconManager.prototype.postAnnotationOnSuccess = function(newAnnotation) {
    //NOOP: let Annozilla call addAnnotation
    // though it wouldn't be too bad to resolve redundantly below in this case
}
/*
    try {

        if (this._isInsertIconsEnabled()) {
            var aImg = null;

            // try to resolve the annotation
            var range = this.annozilla.resolveAnnotation(newAnnotation);
            if (range) {
                aImg = this._insertIcon(newAnnotation, range);
            }
        }
    } catch (e) {
        Components.classes['@mozilla.org/consoleservice;1'].getService(
            Components.interfaces.nsIConsoleService
        ).logStringMessage(
            "Annozilla: Failure in IconManager.postAnnotationOnSuccess: " + e
        );
    }
}
*/

/**
 * Called when an annotation post fails.
 *
 * @param message string error message
 */
IconManager.prototype.postAnnotationOnFailure = function(message) {}



/**
 * Called when an annotation update succeeds.
 *
 * @param updatedAnnotation nsIAnnotation that has been updated on server
 */
IconManager.prototype.updateAnnotationOnSuccess = function(updatedAnnotation) {

    if (this._isInsertIconsEnabled()) {

        var range = this.annozilla.resolveAnnotation(updatedAnnotation);
        this._insertAnnotationBefore(updatedAnnotation, range);
        /*
        // update the type, if there is an icon, and it is displaying
        var img = this.icons[updatedAnnotation.annotationURI];
        if (img && img.parentNode) {
            var type = updatedAnnotation.data.getProperty(
                Annozilla.RDF_NS,
                "type"
            );

            if(type) type = type.substring(type.indexOf("#") + 1);
            else type = "Annotation";

            img.setAttribute("src", IconManager._getIconURI(type));
            img.setAttribute("alt", type);
        }
        */

    }
}

/**
 * Called when an annotation update fails.
 *
 * @param url string url of the annotation that could not be updated
 * @param message string error message
 */
IconManager.prototype.updateAnnotationOnFailure = function(url, message) {}


/**
 * Called when an annotation is successfully deleted.
 *
 * @param url string url of deleted annotation
 */
IconManager.prototype.deleteAnnotationOnSuccess = function(uri) {

    if (this._isInsertIconsEnabled()) {

        this._removeHighlight(uri);
        
        var img = this.icons[uri];
        if (img) {
            // if there is an img, remove it
            this._removeIcon(img);
        }
    }
}


IconManager.prototype._removeHighlight = function (uri) {
    //hey, annozilla.annotation[id] doesn't exist anymore?
    // we have to assume it's for the current window
    //remove associated spans 
    // can we know when to skip this? what if we're not highlighting?
    var spans = window._content.document.getElementsByTagName('span');
    for (var i=0; i < spans.length; i++) {
        var span = spans[i];
        var parent = span.parentNode;
        if (span.getAttribute('annotation-id') == uri) {
            if (span.parentNode) {
                while (span.hasChildNodes()) {
                    var child = span.removeChild(span.firstChild);
                    parent.insertBefore(child, span);
                }
                parent.removeChild(span);
            }
        }
    }
}

/**
 * Called when an annotation delete fails.
 *
 * @param url string url of the annotation that could not be deleted
 * @param message string error message
 */
IconManager.prototype.deleteAnnotationOnFailure = function(url, message) {}

/**
 * Called when the replies to an annotation have been successfully
 * loaded.
 * 
 * @param annotation nsIAnnotation for which replies were requested
 * @param repliesEnum nsISimpleEnumerator containing nsIAnnotation replies
 */
IconManager.prototype.getRepliesOnSuccess = function(annotation, repliesCount, replies) {}

/**
 * Called when getting the replies to an annotation fails.
 *
 * @param annotation nsIAnnotation for which replies were requested
 * @param message string error message
 */
IconManager.prototype.getRepliesOnFailure = function(annotation, message) {}

/**
 * Called an annotation has successfully been replied to.
 * 
 * @param original nsIAnnotation that existed originally
 * @param reply nsIAnnotation reply created
 */
IconManager.prototype.replyToAnnotationOnSuccess = function(original, reply) {}

/**
 * Called when replying to an annotation fails.
 *
 * @param original nsIAnnotation to which we attempted to reply
 * @param message string error message
 */
IconManager.prototype.replyToAnnotationOnFailure = function(original, message) {}

/**
 * Called after a loadAnnotationBody request
 * when the body text of the annotation has been successfully loaded.
 *
 * @param annotation nsIAnnotation with updated body text
 * @param bodyText string text of the annotation's body
 */
IconManager.prototype.loadAnnotationBodyOnSuccess = function(annotation, bodyText) {}

/**
 * Called after a loadAnnotationBody request when
 * the body text fails to load.
 *
 * @param annotation nsIAnnotation that body was requested for (unmodified)
 * @param message string error message
 */
IconManager.prototype.loadAnnotationBodyOnFailure = function(annotation, message) {}

IconManager.prototype.iconClick = function(annotation) {
    if(annotation) this.annozilla.update(annotation.annotationURI);
}

IconManager.prototype.addIconForEditorWindow = function(win, fragment){
    var range = this.annozilla.resolveFragment(fragment);
    if(!range) throw "resolveFragment failed for fragment '" + fragment + "'";
    var icon  = this._insertIcon(null, range);
    var im = this;
    win.addEventListener(
        "unload",
        function(e){
            if(win.closed) im._removeIcon(icon);
        },
        false
    );
};

/********************
 * PRIVATE FUNCTIONS
 ********************/

IconManager._hasRequiredComponents = function() {
    return true;
}

// Add the icon icon to this.icons.
IconManager.prototype._addIconToIcons = function(icon, annotationURI){
    this.icons[annotationURI] = icon;
};

IconManager.prototype._iconForAnnotation = function(annotation) {
    // create the icon
    var img = window._content.document.createElementNS(
        Annozilla.XHTML_NS,
        "img"
    );
    if(annotation){
        var type = annotation.data.getProperty(Annozilla.RDF_NS, "type");

        if(type){
            type = type.substring(type.indexOf("#") + 1);
        }else{
            type = "Annotation";
        }

        img.setAttribute("src", IconManager._getIconURI(type));
        img.setAttribute("class", "annozillaAnnotationIcon"); 
    }else{
        img.setAttribute("src", "chrome://annozilla/content/icons/newAnnotation.png");
        img.setAttribute("class", "annozillaUnsavedAnnotationIcon"); 
    }

    return img;
}

IconManager.prototype._insertIconBefore = function(annotation, range, icon) {
    // icon is now added to icons in _insertIcon.
//  if(annotation) this.icons[annotation.annotationURI] = icon;
//  else this.temporaryIcons.push({ icon: icon, location: range});

    // mark the element to be ignored by XPointer resolution
    xptrMarkElement(icon);

    IconManager._insertNode(range, icon);

    var that = this;

    icon.addEventListener(
        "click",
        function(e){
            //prevent link action when the icon is inside the anchor <a> area.
            e.preventDefault();
            that.iconClick(annotation);
        },
        false
    );

    return icon;
}

/**
 *  Inserts highlighting span and/or icon image
 *  Unless such elements are already found with associated annotation-id attribute
 *  - this allows for cached page or reload of sbar but requires page
 *  to be reloaded or elements removed to refresh 
 *  (annotation-id attribute may be useful for removal/edit)
 *
 *  - ideally we'd want some sort of checksum to make sure annotation length
 *    hasn't changed?
 *
 *  Also note:
 *   annotation.data "type" property of "type#subtype" 
 *     allows new type to share a particular icon subtype style? (experimental)
 *
 *   new icon given spanlength attribute to allow for proper nesting
 *
 */
IconManager.prototype._insertAnnotationBefore = function(annotation, range) {
    if (!range) throw "a range is required for _insertAnnotationBefore";

    var uri = annotation.annotationURI;
    var type = annotation.data.getProperty(IconManager.RDF_NS, "type");
    type = type?type.substring(type.indexOf("#") + 1):"Annotation";

    //first see whether there are already icons or spans for this 
    var img;
    //the parent of the start container may be an existing highlight span/div
    var span = range.startContainer.parentNode;

    //highlights insert page annotations just under html
    // so instead of searching up the tree, need to search forward
    // starting with first child of the html node
    //also check page annotation body for a stylesheet directive
    var seekForward = false;
    if (range.startContainer.nodeName.toLowerCase() == "html") {
        //dump("[Page Annot]");
        span = range.startContainer.firstChild;
        seekForward = true;

        //loading body might take too long for remote server to be useful here
        // but it works for localserver
        // see loadAnnotationBody[OnSuccess]?
        var body = annotation.getBodyText();
        try { //don't choke if we can't get body 
            var matches = body.match(/Use-Stylesheet: ([^<]*)/g);
            if (matches) {
                for (var i = 0; i < matches.length; i++) {
                    dump("* Use-Stylesheet: "+matches[i].substr(16)+"\n");
                    addStyleSheet(window.content.document, matches[i].substr(16));
                }
            }
        } catch (e) {
            dump("Trouble with annotation body\n");
        }
    } 

    //and if we're just inside a document, we'd better move inside
    // this can happen when an annotation was created without a selection
    // annotations really ought to be restricted to body elements?
    if (span.nodeType == span.DOCUMENT_NODE) {
        span = range.startContainer.firstChild;
    }

    // highlighted spans might be nested
    while (span) {
        if (span.nodeType == span.ELEMENT_NODE) {
            //&& (span.hasAttribute("_dom_ignore_element_"))) {
            // first, get above wrapped elements
            if ((span.hasAttribute("_dom_ignore_element_"))
                && (span.getAttribute("annotation-id") == uri)
                && ((span.nodeName == "span")
                    || (span.nodeName == "div"))) {
                //typically, we assume it is our span or div
                // but if seekForward it may be an img so we need to check
                break;
            } else {
                if (seekForward) {
                    //page annotations are inserted one after another
                    span = span.nextSibling;
                } else {
                    // keep going up
                    //  assuming for now that icons don't appear without spans,
                    //  else we might have to look at previousSibling
                    span = span.parentNode; 
                }
                
            }
        } else {
            span = null;
        }
    }        

    // highlight first because _insertNode can alter dom path to end of range
    if (this.annozilla.prefs.getPref("insertHighlight")) {
        if (span) {
            //update existing span
            span.setAttribute("annotation-type", type);
        } else {
            //insert span/div 
            //  with class, annotation-type, annotation-id, segment attributes
            doHighlight(range, type, annotation);
            //move back one to insert icon before span
            if (range.startOffset > 0) {
                range.setStart(range.startContainer,range.startOffset-1);
            }
        } 
    }

    if (this.insertIcons) {
        // now look where there should be an icon
        if (span) {
            img = span.previousSibling;
        } else {
            img = range.startContainer.previousSibling;
        }
        
        while (img) {
            if ((img.nodeType == img.ELEMENT_NODE)
                && (img.nodeName == "img")
                && (img.getAttribute("annotation-id") == uri)) {
               //update existing icon
               img.setAttribute("annotation-type", type);
               img.setAttribute("src", IconManager._getIconURI(type));
               img.setAttribute("alt", type);               
               return img; 
             } else {
                 //might be a footnote or another img?
                 img = img.previousSibling;
             }
        }
        // else... new icon should go before existing span
        if (span) {
            range.setStartBefore(span);
        }
        
        img = this._iconForAnnotation(annotation);

        // mark the element to be ignored by XPointer resolution
        xptrMarkElement(img);

        img.setAttribute("annotation-id", uri);
        //allow selection for CSS, XBL (experimental)
        img.setAttribute("annotation-type", type); 
        // spanlength allows highlight to determine proper nesting
        img.setAttribute("spanlength", range.toString().length);

        img = this._insertIconBefore(annotation, range, img);
        
    };
    return img;

}

// This function adds an icon for annotation annotation at the start of range
// range within the current document (window._content.document), or adds a
// temporary icon (for an as yet unsaved annotation) if annotation is null.
IconManager.prototype._insertIcon = function(annotation, range) {
    var icon;
    if(range == null) {
        // Insert the icon at the start of the document.
        // (this is not used (or tested) because the two places that call
        // _insertIcon both check that range is not null).
        icon = this._insertAnnotationBefore(
            annotation,
            window._content.document.body.firstChild
        );
    } else {
        icon = this._insertAnnotationBefore(annotation, range);
    }
    if(annotation) { this._addIconToIcons(icon, annotation.annotationURI); }
    return icon;
}

IconManager._getIconURI = function(type) {     
  switch(type) {
  case 'Comment':
    return 'chrome://annozilla/content/icons/annot.png';

  case 'Advice':
    return 'chrome://annozilla/content/icons/advice.png';

  case 'Example':
    return 'chrome://annozilla/content/icons/example.png';

  case 'Explanation':
    return 'chrome://annozilla/content/icons/explanation.png';

  case 'Question':
    return 'chrome://annozilla/content/icons/question.png';

  case 'SeeAlso':
    return 'chrome://annozilla/content/icons/seealso.png';

  case 'Change':
    return 'chrome://annozilla/content/icons/change.png';

  default:
    return 'chrome://annozilla/content/icons/annot.png';
  }
}



/**
 * Inserts a node into the Document or DocumentFragment at the start
 * of the Range. If the container is a Text node, this will be split
 * at the start of the Range (as if the Text node's splitText method
 * was performed at the insertion point) and the insertion will occur
 * between the two resulting Text nodes. Adjacent Text nodes will not
 * be automatically merged. If the node to be inserted is a
 * DocumentFragment node, the children will be inserted rather than
 * the DocumentFragment node itself.
 *
 * @param range DOM Range Object
 * @param node DOM Node to insert
 */
IconManager._insertNode = function(range, node_to_insert) {

    var console = Components.classes['@mozilla.org/consoleservice;1']
        .getService(Components.interfaces.nsIConsoleService);
    try {
        var n = range.startContainer; // just choose a node for now

        // evil (overwrites existing style):
        n.setAttribute ("style", "-moz-binding: url('chrome://annozilla/content/bindings/annozillaBindings.xml#icon')");

        // Don't seem to be able to perform this instantaneously
        // (binding document doesn't appear to load synchronously)
        window.setTimeout (
             function (node) {
                 // Pick URL and icon arbitrarily for now
                 node.newIcon ("about:blank", "chrome://annozilla/content/icons/annot.png");
             },
             100,
             n.wrappedJSObject);
    } catch (e) {
        console.logStringMessage ("XBLIconManager " + e);
    }
/*
  var ref_node;   // node beyond insertion point
  var parent;     // parent node
  
  if (range.startContainer.nodeType == node_to_insert.TEXT_NODE || 
      range.startContainer.nodeType == node_to_insert.CDATA_SECTION_NODE) {
    ref_node = range.startContainer.splitText(range.startOffset);
    
    range.startContainer.parentNode.insertBefore(node_to_insert, ref_node);
    return;
  }    
  
  // if the insertion point is at the end of the startContainer node just 
  // append the node to the startContainer node.
  if ( range.startOffset == range.startContainer.childNodes.length) {
    range.startContainer.appendChild( node_to_insert );
  }
  // otherwise, insertion point is not at end of startContainer
  else {
    range.startContainer.insertBefore( node_to_insert,
				       range.startContainer.childNodes.item(range.startOffset));
  }
*/
}

IconManager.prototype.removeIcons = function() {
    for (uri in this.icons) {
        var img = this.icons[uri];
        if (img) this._removeIcon(img);
    }
    this.icons = {};
}

IconManager.prototype._removeIcon = function(icon) {
    if (icon.parentNode) {
        icon.parentNode.removeChild(icon);
    }
}

/**
 * Resolve each annotation before calling _insertAnnotation 
 * Don't sort first - it may have been useful to insert in reverse order 
 *   so that inserted icons wouldn't invalidate xpointers, but inserting
 *    spans for highlighting we have to resolve them each time anyway 
 *
 * @param annotations collection to insert, 
 *                     resolved xpointer of each returned as range property
 */
/* THIS ISN'T CALLED ANYMORE
IconManager.prototype._insertAnnotations = function(annotations) {
    dump("[_insertAnnotations]\n");
    for (uri in annotations) {
        dump(".");
	var annotationObj = annotations[uri];
	var annot = annotationObj.annot;
	if (! annot.data) { continue; } // ignore bad annotation

        // previous annotations may have changed dom
        var range = xptrResolveAnnotation(annot, window._content.document);
        annotationObj.range = range;

	var img = null;
	if (range) {
	    img = this._insertAnnotation(annot, range);
	}
        annotationObj.img = img;
    }
    dump("\n");
}
*/

/*
 * Old version
 *
 * annotations are sorted by their positions before inserting.
 * written by K.Tsujino

IconManager.prototype._insertAnnotations = function(annotations) {

    var alist = new Array();
    for(uri in annotations) { alist.push(annotations[uri]); }
    alist.sort( function(x,y) {
                    if( y.range && y.range.startOffset && x.range && x.range.startOffset )
                      return y.range.startOffset - x.range.startOffset;
                    return 0;
                } );

    for( var i=0 ; i<alist.length ; ++i ) {
        var annotationObj = alist[i];

	var annot = annotationObj.annot;
	if (! annot.data) {
	    // ignore it
	    continue;
	}

        var range = annotationObj.range;
	var img = null;
	if (range) {
	    img = this._insertIcon(annot, range);
	}
        // Icons are now added to this.icons in _insertIcon.
//      this.icons[alist[i].annot.annotationURI] = img;
    }
}
 */

/* THIS ISN'T CALLED ANYMORE, PROCESSING NOW HANGLED THROUGH ANNOZILLA
IconManager.prototype._processAnnotationsWithAutoLoadInfo = function(state, annotations) {
   
    if (state == Annozilla.PAGE_LOADING) {
        return;
    }
    else if (state == Annozilla.PAGE_COMPLETED) {
	var currentAnnotationsBatch = {};
	
	for (var i = 0; i < annotations.length; i++) {
	    var current = annotations[i];
	
	    if (! current.data) {
		this.annozilla._debugBadAnnotation(current);
		continue;
	    }
	    
	    var range = this.annozilla.resolveAnnotation(current);
	    
	    // add to the current batch
            IconManager._addAnnotationToContainer(
                current,
                range,
                null,
                currentAnnotationsBatch
            );
	}
      
        this.annozilla.addAnnotations(currentAnnotationsBatch);
    }
    
    // otherwise, the page failed to load, so we ignore annotations
    
}

IconManager._addAnnotationToContainer = function(
    aAnnotation,
    aRange,
    aImg,
    aContainer
){
    aContainer[aAnnotation.annotationURI] = {
        annot:    aAnnotation,
        range: aRange,
        img:      aImg
    };
}
*/

// _isInsertIcons[OrHighlight]Enabled
IconManager.prototype._isInsertIconsEnabled = function() {
    return this.insertIcons || this.insertHighlight;
    //this.annozilla.prefs.getPref("insertHighlight");
}

/**
 * Called during load (by Annozilla..addAnnotations/_AddNext)  
 *   with xpointer resolution already taken care of
 */
IconManager.prototype.addAnnotation = function(annotation, range) {
    if (this._isInsertIconsEnabled()) {    
        var img = this._insertAnnotationBefore(annotation, range);
        //this.annotations[annotation.annotationURI].img = img;
        this._addIconToIcons(img, annotation.annotationURI);
    }
}

